// Author      : Romain Testylier
// Version     :
// Copyright   : 
// Description : Example of reachability analysis using
//               hybridization on a 7 dimensional system.
//============================================================================

#include <iostream>
#include <string>
#include <fstream>
#include <iomanip>
#include <stdlib.h>
#include <time.h>

// include from the library
#include <MatlabExporter.h>
#include <utils.h>
#include <ReachHybridization.h>

using namespace std;
using namespace nltool;
using namespace arma;

// the derivatives of the dynamical system 

double df(const int x, const stdVector &pt, void* data) {

	double k1 = 1.4;
	double k2 = 0.9;
	double k3 = 2.5;
	double k4 = 1.5;
	double k5 = 0.6;
	double k6 = 0.8;
	double k7 = 2.0;
	double k8 = 1.3;
	double k9 = 0.7;
	double k10 = 1.0;
	double k11 = 0.3;
	double k12 = 3.1;
	double k13 = 1.8;
	double k14 = 1.5;

	switch (x) {
	case 0:
		return k1 * pt[2] - k2 * pt[0];
		break;
	case 1:
		return k3 * pt[4] - k4 * pt[1];
		break;
	case 2:
		return k5 * pt[6] - k6 * pt[2] * pt[1];
		break;
	case 3:
		return k7 - k8 * pt[3] * pt[2];
		break;
	case 4:
		return k9 * pt[0] - k10 * pt[3] * pt[4];
		break;
	case 5:
		return k11 * pt[0] - k12 * pt[5];
		break;
	case 6:
		return k13 * pt[5] - k14 * pt[6] * pt[1];
		break;
	}
}

// The following C function "hess" specifies the Hessian matrix of the derivative functions 
// This function returns the partial derivative 
// The integer x is the index of the above derivative function "df"
// The integers dx and dy are the indices of the variables over the partial derivatives are defined
// data is for additional information, unused in this example
double hess(const int x, const stdVector &pt, const int dx, const int dy,
		void* data) {
	double k1 = 1.4;
	double k2 = 0.9;
	double k3 = 2.5;
	double k4 = 1.5;
	double k5 = 0.6;
	double k6 = 0.8;
	double k7 = 2.0;
	double k8 = 1.3;
	double k9 = 0.7;
	double k10 = 1.0;
	double k11 = 0.3;
	double k12 = 3.1;
	double k13 = 1.8;
	double k14 = 1.5;
	if ((x == 2) && (((dx == 2) && (dy == 1)) || ((dx == 1) && (dy == 2))))
		return -k6;
	if ((x == 3) && (((dx == 2) && (dy == 3)) || ((dx == 3) && (dy == 2))))
		return -k8;
	if ((x == 4) && (((dx == 3) && (dy == 4)) || ((dx == 4) && (dy == 3))))
		return -k10;
	if ((x == 6) && (((dx == 6) && (dy == 1)) || ((dx == 1) && (dy == 6))))
		return -k14;
	return 0;

}


// The following main function contains the basic function calls to perform the computation 
// of the reachable sets for a number of iterations

int main(int argc, char **argv) {
	int dim = 7; // dimension of the dynamical system

	double dt = 0.007; // time step
	int nbIter = 2000; // number of the reachability computation iterations
	double error = 0.025; // desired hybridization error bound 
	Chrono chrono; // for estimating computational time

	///////
        // Defining the initial set which is octogonal
        // variable "translation" defines the centroid of the initial set
	Vector translation(dim);
	translation[0] = 1.2;
	translation[1] = 1.05;
	translation[2] = 1.5;
	translation[3] = 2.4;
	translation[4] = 1.0;
	translation[5] = 0.1;
	translation[6] = 0.45;

	Template tmpTemplate;
	Vector tmpCoef;

        // variable "radius" defines the radius if the circumsphere of the initial set
	double radius;

        // creating the templates (constraints) of the initial set
	createOctagonalSet(dim, radius, translation, tmpTemplate, tmpCoef);

        // creating the initial set from the templates
	Hpolyhedron initialSet(dim, tmpTemplate, tmpCoef);


        ///////
	// Creating a dynamical system with the above derivative functions
	PointerSystem ptrSys(dim, df, NULL, hess, NULL, false);

	///////
	// Creating an instance of reachability computation using the hybridization method
	ReachHybridization reachHybridization(
			static_cast<DynamicalSystem*>(&ptrSys), initialSet, error, dt);
	
	
	chrono.start(); // start the clock to count the computation time
	
	///////
	// Performing reachability computation
	reachHybridization.reach(nbIter);

	double time = chrono.stop(); // getting the computation time
	cout << "duration: " << time << " s" << endl;


	///////
	// Getting the reachable set as a list of polyhedra, for visualization purposes
	vector<Hpolyhedron> result = reachHybridization.getReachabilityResult();

	///////
	// Getting the hybridization domains, for visualization purposes
	vector<Hpolyhedron> approximationDomains =
			reachHybridization.getDomainsCreated();


	// exporting the result in a Matlab script
	std::vector<int> projection(2); // specifiy the projected variables
	stdVector color(3, 0);
	color[0] = 1;
	
	// "variable "projection" contains the projection axes
	projection[0] = 4;
	projection[1] = 6;

	///////
	// output the result to a file named "dictyos_hybridi_x"
	ostringstream oss;
	oss << "dictyos_hybridi_x" << projection[0] << "x" << projection[1] << ".m";
	string filename = oss.str();

	///////
	// specifying the frequency of writing the result to the file (every 8 iterations in this example)
	int displayEvery = 8;

	///////
	// creating the templates for the projections that are going to be visualized
	stdVector exporterBl, exporterBu = stdVector(dim);
	Template approxTemplate = createTemplateApproximation(dim, 8, projection[0],
			projection[1]);
	Template approxTemplateDomains = createTemplateApproximation(dim, 12,
			projection[0], projection[1]);

	///////
	// creating an instance for result export
	MatlabExporter exporter(dim, filename, exporterBl, exporterBu, projection);

	///////
	// Writing the hybridization domains to the file
	exporter.setApproximationTemplate(approxTemplateDomains);
	for (unsigned i = 0; i < approximationDomains.size(); i += displayEvery)
		exporter.save(approximationDomains[i]);
	exporter.findBounds(approximationDomains);

	// writing the reachable state sets to the file
	exporter.setApproximationTemplate(approxTemplate);
	for (unsigned i = 0; i < result.size(); i += displayEvery)
		exporter.save(result[i], color);

	cout << "DONE!!" << endl;
	return 0;
}
