// Author      : Romain Testylier
// Version     :
// Copyright   : 
// Description : Example of reachability analysis using
//               Bernstein expansion on a 7 dimensional system.
//============================================================================

#include <iostream>
#include <string>
#include <fstream>
#include <iomanip>
#include <stdlib.h>
#include <time.h>
//
#include <MatlabExporter.h>
#include <utils.h>
#include <ReachPolynomial.h>

using namespace std;
using namespace nltool;
using namespace arma;

// system description from string

string dyctios_str() {

	double k1 = 1.4;
	double k2 = 0.9;
	double k3 = 2.5;
	double k4 = 1.5;
	double k5 = 0.6;
	double k6 = 0.8;
	double k7 = 2.0;
	double k8 = 1.3;
	double k9 = 0.7;
	double k10 = 1.0;
	double k11 = 0.3;
	double k12 = 3.1;
	double k13 = 1.8;
	double k14 = 1.5;

	ostringstream str;
	str << "x0' = " << k1 << "*x2 - " << k2 << "*x0" << endl;
	str << "x1' = " << k3 << "*x4 - " << k4 << "*x1" << endl;
	str << "x2' = " << k5 << "*x6 - " << k6 << "*x2*x1" << endl;
	str << "x3' = " << k7 << " - " << k8 << "*x3*x2" << endl;
	str << "x4' = " << k9 << "*x0 - " << k10 << "*x3*x4" << endl;
	str << "x5' = " << k11 << "*x0 - " << k12 << "*x5" << endl;
	str << "x6' = " << k13 << "*x5 - " << k14 << "*x6*x1";

	return str.str();
}

int main(int argc, char **argv) {
	int dim = 7; // dimension of the dynamical system
	Number dt = 0.028; // time step
	int nbIter = 500; // number of the reachability computation iterations
	Chrono chrono; // for estimating computational time

	///////
        // Defining the initial set which is octogonal
        // variable "translation" defines the centroid of the initial set
	Vector translation(dim);
	translation[0] = 1.2;
	translation[1] = 1.05;
	translation[2] = 1.5;
	translation[3] = 2.4;
	translation[4] = 1.0;
	translation[5] = 0.1;
	translation[6] = 0.45;



	Template tmpTemplate;
	Vector tmpCoef;

	// variable "radius" defines the radius if the circumsphere of the initial set
	Number radius = 0.0001;

	// creating the rectangular initial set from the templates, by defining the two vectors of bounding values
	stdVector bl = arma::conv_to < stdVector > ::from(translation - radius);
	stdVector bu = arma::conv_to < stdVector > ::from(translation + radius);
	Box tmpBox(bl, bu);
	Hpolyhedron initialSet(tmpBox);

	///////
	// Creating a dynamical system 
	string systemStr = dyctios_str();
	PolynomialSystem polySys(dim, systemStr);
	polySys = polySys.getDiscretizedSystem(dt);

	ReachPolynomial reachPoly(polySys);
	vector<Hpolyhedron> result;
	result.push_back(initialSet);

	chrono.start();
        ///////
	// Performing reachability computation
	for (unsigned i = 0; i < nbIter; i++) {
		cout << "iteration " << i << endl;
		Hpolyhedron res = reachPoly.reach(result[result.size() - 1]);
		result.push_back(res);

	}
	Number time = chrono.stop();
	cout << "duration: " << time << " s" << endl;

	// exporting results
	std::vector<int> projection(2); // specifiy the projected variables
	projection[0] = 4;
	projection[1] = 6;

	///////
	// output the result to a file named "dictyos_bern_x.m"
	ostringstream oss;
	oss << "dictyos_bern_x" << projection[0] << "x" << projection[1] << ".m";
	string filename = oss.str();
	stdVector color(3, 0);
	color[1] = 1;

	
	///////
	// creating the templates for the projections that are going to be visualized
	stdVector exporterBl, exporterBu = stdVector(dim);
	Template approxTemplate = createTemplateApproximation(dim, 4, projection[0],
			projection[1]);

	///////
	// creating an instance for result export
	MatlabExporter exporter(dim, filename, exporterBl, exporterBu, projection);

	//exporter.setApproximationTemplate(approxTemplate);
	exporter.findBounds(result);

	///////
	// specifying the frequency of writing the result to the file (every 8 iterations in this example)
	int displayEvery = 3;

	// writing the reachable state sets to the file
	for (unsigned i = 0; i < result.size(); i += displayEvery) {
		exporter.save(result[i]);
	}

	cout << "DONE!!" << endl;
	return 0;
}
